class SaladPyException(Exception):
    """
    General SaladPy exception.
    Packs together all the possible errors Salad API could give
    """

    def __init__(self, message: str) -> None:
        self.message = message

    def __repr__(self) -> str:
        return '<SaladPyException(message="{0}")>'.format(
            self.message
        )


import json

from typing import Optional
from pathlib import Path
from aiohttp import ClientSession
from types import SimpleNamespace

from .exceptions import *
from .methods import Methods


class SaladClient(Methods):
    V1_API_URL = "https://app-api.salad.com/api/v1/"
    V2_API_URL = "https://app-api.salad.com/api/v2/"
    AUTH_API_URL = "https://app-api.salad.com/auth/"

    def __init__(
        self,
        cachePath: Optional[Path] = None,
        *args,
        **kwargs,
    ):
        self.cachePath = cachePath
        self.cached = False
        self.http = ClientSession(*args, **kwargs)
        if self.cachePath is not None:
            if self.cachePath.exists():
                if self.cachePath.stat().st_size > 0:  # File is not empty
                    self.http.cookie_jar.load(self.cachePath)
                    self.cached = True
                    
    @property
    def cookies(self):
        return {cookie.key: cookie.value for cookie in list(self.http.cookie_jar) }
                            
    async def close(self):
        """
        Closes aiohttp `ClientSession` safely
        Using any methods after calling `close` will not work
        """
        await self.http.close()

    async def login(self, email: str):
        """Gets the necessary authentication tokens required for API requests

        Args:
            email (str): Salad E-Mail

        Raises:
            SaladPyException: An error during Email Verification

        Returns:
            Callable: A function representing the `_verify` method, takes in otp as argument
        """
        if not self.cached:
            headers = {"Content-Type": "application/json"}
            payload = json.dumps({"email": email, "termsAccepted": True})
            async with self.http.post(
                f"{self.V2_API_URL}authentication-sessions",
                data=payload,
                headers=headers
            ) as r:
                try:
                    assert r.status == 204
                    # Ensure that the `sInitToken` cookie is correctly applied to the session
                    # If not done correctly, OTP verification will not work
                    self.http.cookie_jar.update_cookies(r.cookies)
                except AssertionError:
                    text = await r.text()
                    raise SaladPyException(
                        f"Error {r.status} during email verification! {text}"
                    )
                else:
                    return self._verify

    async def _verify(self, otp: str):
        """Verify using OTP sent to Salad E-Mail

        Args:
            otp (str): The OTP given by the user

        Raises:
            SaladPyException: An error during OTP verification

        Returns:
            Optional[Bool]: True if no errors occured
        """
        if not self.cached:
            headers = {
                "Content-Type": "application/json",
            }
            payload = {"passcode": str(otp)}
            async with self.http.post(
                f"{self.V2_API_URL}authentication-sessions/verification",
                json=payload,
                headers=headers,
                cookies=self.cookies,
            ) as r:
                try:
                    cookies = list(self.http.cookie_jar)
                    assert r.status == 204
                    self.http.cookie_jar.update_cookies(r.cookies)
                    if self.cachePath is not None:
                        self.cachePath.touch(exist_ok=True)
                        self.http.cookie_jar.save(self.cachePath)
                except AssertionError:
                    text = await r.json()
                    raise SaladPyException(
                        f"Error {r.status} during OTP verification! {text}"
                    )
                else:
                    return True

    async def _req(self, api_url : str, method: str, endpoint: str, params: Optional[dict] = None):
        async with self.http.request(
            method, f"{api_url}{endpoint}", params=params
        ) as r:
            text = await r.text()
            try:
                assert r.status == 200
            except AssertionError:
                if (
                    "try refresh token" in text
                ):  # DO NOT convert to JSON as it's not always guaranteed to be JSON
                    # Refresh token
                    print("Token expired, refreshing...")
                    await self.refresh_token()
                else:
                    return r.status
            else:
                return text

    async def refresh_token(self):
        #TODO: Test refresh token
        self.temp_cookie_backup = self.cookies
        cookies = {"sRefreshToken": self.cookies["sRefreshToken"], "sIdRefreshToken": self.cookies["sIdRefreshToken"]}
        self.http.cookie_jar.clear()
        async with self.http.post(f"{self.AUTH_API_URL}session/refresh", cookies=cookies) as r:
            print(r.status)
            text = await r.text()
            print(text)
            if r.status == 200:
                self.http.cookie_jar.update_cookies(r.cookies)
                self.http.cookie_jar.save(self.cachePath)

    async def _get(self, api_url: str, endpoint: str, params: Optional[dict] = None):
        resp = await self._req(api_url, "GET", endpoint, params)
        if resp == 404:
            return resp
        return json.loads(resp)

    async def _post(self, api_url: str, endpoint: str, params: Optional[dict] = None):
        resp = await self._req(api_url, "POST", endpoint, params)
        if resp == 404:
            return resp
        return json.loads(resp)



from typing import NamedTuple

class ProfileResponse(NamedTuple):
    email: str
    extensions: dict
    id: str
    lastSeenApplicationVersion: str
    termsVersion: str
    termsAcceptedAt: str
    username: str
    viewedReferralOnboarding: bool

class BalanceResponse(NamedTuple):
    currentBalance: float
    lifetimeBalance: float
    
class ReferralDefinition(NamedTuple):
    balanceThreshold: float
    bonusRate: float
    referrerBonus: float
    
class ReferralsResponse(NamedTuple):
    code: str
    dateEntered: str
    earnedBalance: float
    refereeId: str
    referralDefinition: ReferralDefinition
    referrerId: str
    
class RedemptionsResponse(NamedTuple):
    code: str
    id: str
    name: str
    price: float
    status: str
    timestamp: str

class AvatarsResponse(NamedTuple):
    description: str
    id: str
    imageUrl: str
    name: str

class SelectedAvatarResponse(AvatarsResponse):
    pass


"""A class containing all the User API endpoints"""
from typing import List, Dict, Union

from .types import *


class UserMethods(BaseClient):
    """A class containing all the User API endpoints"""

    async def profile(self) -> ProfileResponse:
        """Gets a User's Profile

        Returns:
            ProfileResponse: Info about the User's profile
        """
        return ProfileResponse(**await self._get(self.V1_API_URL, "profile"))

    async def balance(self) -> BalanceResponse:
        """Gets a User's Balance

        Returns:
            BalanceResponse: The user's currentBalance and lifetimeBalance
        """
        return BalanceResponse(**await self._get(self.V1_API_URL, "profile/balance"))

    async def xp(self) -> int:
        """Gets a User's xp

        Returns:
            int: Lifetime Xp
        """
        return (await self._get(self.V1_API_URL, "profile/xp"))["lifetimeXp"]

    async def referral_code(self) -> str:
        """Gets a User's referral code

        Returns:
            str: User's referral code
        """
        return (await self._get(self.V1_API_URL, "profile/referral-code"))["code"]

    async def referrals(self) -> List[ReferralsResponse]:
        """Gets a list of users that have used a User's code

        Returns:
            List[ReferralsResponse]: List of users that have used your referral code
        """
        result = await self._get(self.V1_API_URL, "profile/referrals")
        response = []
        for resp in result:
            ref_def = ReferralDefinition(**resp["referralDefinition"])
            resp.pop("referralDefinition", None)
            response.append(ReferralsResponse(referralDefinition=ref_def, **resp))
        return response

    async def redemptions(self) -> List[RedemptionsResponse]:
        """Gets a User's Redemptions

        Returns:
            List[RedemptionsResponse]: A list of a user's redemptions, will be empty if there are none
        """
        result = await self._get(self.V1_API_URL, "reward-vault")
        response = []
        for resp in result:
            if "code" not in resp:
                resp["code"] = ""
            response.append(RedemptionsResponse(**resp))
        return response

    async def one_day_earning_history(self) -> Dict[str, float]:
        """Gets a User's 1 day earning history

        Returns:
            Dict[str, float]: 1 Day earning history, where the key is a time and the value is a float, empty if no earning history
        """
        return await self._get(self.V2_API_URL, "reports/1-day-earning-history")  # type: ignore

    async def seven_day_earning_history(self) -> Dict[str, float]:
        """Gets a User's 7 day earning history

        Returns:
            Dict[str, float]: 7 Day earning history, where the key is a time and the value is a float, empty if no earning history
        """
        return await self._get(self.V2_API_URL, "reports/7-day-earning-history")  # type: ignore

    async def thirty_day_earning_history(self) -> Dict[str, float]:
        """Gets a User's 30 day earning history

        Returns:
            Dict[str, float]: 30 Day earning history, where the key is a time and the value is a float, empty if no earning history
        """
        return await self._get(self.V2_API_URL, "reports/30-day-earning-history")  # type: ignore

    async def avatars(self) -> List[AvatarsResponse]:
        """Gets a User's unlocked avatars

        Returns:
            List[AvatarsResponse]: List of unlocked avatars
        """
        return [AvatarsResponse(**resp) for resp in await self._get(self.V2_API_URL, "avatars")]  # type: ignore

    async def selected_avatar(self) -> Union[SelectedAvatarResponse, None]:
        """Gets a User's currently selected avatar

        Returns:
            SelectedAvatarResponse: Currently selected avatar
            None: If using the default avatar
        """
        resp = await self._get(self.V2_API_URL, "avatars/selected")  # type: ignore
        if resp == 404:
            return None
        return SelectedAvatarResponse(**resp)

    async def bonuses(self) -> Union[list, None]:
        """Gets a User's unclaimed bonuses

        Returns:
        TODO: Test what a list of bonuses would look like
            List: List of unclaimed bonuses
            None: If there are no unclaimed bonuses
        """
        resp = await self._get(self.V2_API_URL, "bonuses")  # type: ignore
        if resp == []:
            return None
        return resp



"""A base version of SaladClient used for type hinting"""
from typing import Optional


class BaseClient:
    """A base version of SaladClient used for type hinting"""
    V1_API_URL = "https://app-api.salad.com/api/v1/"
    V2_API_URL = "https://app-api.salad.com/api/v2/"
    AUTH_API_URL = "https://app-api.salad.com/auth/"

    async def _get(self, api_url: str, endpoint: str, params: Optional[dict] = None):
        pass

    async def _post(self, api_url: str, endpoint: str, params: Optional[dict] = None):
        pass

    async def close(self):
        pass

    async def login(self, email: str):
        pass

    async def _verify(self, otp: str):
        pass

    async def refresh_token(self):
        pass

    @property
    def cookies(self):
        pass



from setuptools import setup
import re

requirements = []
with open('requirements.txt') as f:
    requirements = f.read().splitlines()

readme = ''
with open('README.md') as f:
    readme = f.read()

setup(
    name='saladpy',
    author='Coddo-Python',
    url='https://github.com/Coddo-Python/SaladPy',
    project_urls={
        'Github': 'https://github.com/Coddo-Python/SaladPy',
        'Issue tracker': 'https://github.com/Coddo-Python/SaladPy/issues',
        'Docs': 'https://saladpy.gitbook.io/saladpy-docs/'
    },
    version="1.0.2-a.2",
    license='MIT',
    description='A Python wrapper for the Salad Web API',
    long_description=readme,
    packages=["saladpy", "saladpy.methods", "saladpy.methods.types"],
    install_requires=requirements,
    python_requires='>=3.8.0',
    classifiers=[
        'License :: OSI Approved :: MIT License',
        'Intended Audience :: Developers',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Topic :: Internet',
        'Topic :: Software Development :: Libraries',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Utilities',
    ],
)